@@ -303,11 +303,10 @@ class Base:
     #        else:
     #            yield backend.convert(a)
 
-    def make_like(self, *args, **kwargs):
-        op, expr_args = args
+    def make_like(self, op, args, **kwargs):
         if kwargs.pop("simplify", False) is True:
             # Try to simplify the expression again
-            simplified = simplifications.simpleton.simplify(op, expr_args)
+            simplified = simplifications.simpleton.simplify(op, args)
         else:
             simplified = None
         if simplified is not None:
@@ -320,10 +319,10 @@ class Base:
         if 'symbolic' not in kwargs and op in all_operations: kwargs['symbolic'] = self.symbolic
         if simplified is None:
             # Cannot simplify the expression anymore
-            return type(self)(*args, **kwargs)
+            return type(self)(op, args, **kwargs)
         else:
             # The expression is simplified
-            r = type(self)(*(op, simplified.args), **kwargs)
+            r = type(self)(op, simplified.args, **kwargs)
             return r
 
     def _rename(self, new_name):
@@ -518,8 +517,10 @@ class Base:
                 value = '0#{} .. {}'.format(args[0], args[1])
                 return '({})'.format(value) if inner else value
 
-            elif op == 'Concat':
-                return ' .. '.join(map(str, args))
+            elif op in operations.prefix:
+                assert len(args) == 1
+                value = '{}{}'.format(operations.prefix[op], args[0])
+                return '({})'.format(value) if inner and inner_infix_use_par else value
 
             elif op in operations.infix:
                 value = ' {} '.format(operations.infix[op]).join(args)
