@@ -24,6 +24,14 @@ try:
 except ImportError:
     _is_pypy = False
 
+def z3_expr_to_smt2(f, status="unknown", name="benchmark", logic=""):
+      # from https://stackoverflow.com/a/14629021/9719920
+      v = (z3.Ast * 0)()
+      return z3.Z3_benchmark_to_smtlib_string(f.ctx_ref(), name, logic, status, "", 0, v, f.as_ast())
+
+def claripy_solver_to_smt2(s):
+    return s._get_solver().to_smt2()
+
 def _add_memory_pressure(p):
     """
     PyPy's garbage collector is not aware of memory uses happening inside native code. When performing memory-intensive
@@ -46,6 +54,9 @@ solve_count = 0
 
 supports_fp = hasattr(z3, 'fpEQ')
 
+# as of z3 4.8.7.0 this seems to matter
+z3.set_param('rewriter.hi_fp_unspecified', 'true')
+
 #
 # Utility functions
 #
@@ -69,24 +80,14 @@ def _raw_caller(f):
     return raw_caller
 
 def _z3_decl_name_str(ctx, decl):
-    # reimplementation of Z3_get_symbol_string to not try to unicode-decode
-    lib = z3.lib()
-
-    decl_name = lib.Z3_get_decl_name(ctx, decl)
-    err = lib.Z3_get_error_code(ctx)
-    if err != z3.Z3_OK:
-        raise z3.Z3Exception(lib.Z3_get_error_msg(ctx, err))
-
-    symbol_name = lib.Z3_get_symbol_string(ctx, decl_name)
-    err = lib.Z3_get_error_code(ctx)
-    if err != z3.Z3_OK:
-        raise z3.Z3Exception(z3.lib().Z3_get_error_msg(ctx, err))
+    decl_name = z3.Z3_get_decl_name(ctx, decl)
+    symbol_name = z3.Z3_get_symbol_string_bytes(ctx, decl_name)
     return symbol_name
 
 
 class SmartLRUCache(LRUCache):
     def __init__(self, maxsize, getsizeof=None, evict=None):
-        LRUCache.__init__(self, maxsize, getsizeof)
+        LRUCache.__init__(self, maxsize, getsizeof=getsizeof)
         self._evict = evict
 
     def popitem(self):
@@ -121,6 +122,7 @@ class BackendZ3(Backend):
 
         # and the operations
         all_ops = backend_fp_operations | backend_operations if supports_fp else backend_operations
+        all_ops |= backend_strings_operations - {'StrIsDigit'} 
         for o in all_ops - {'BVV', 'BoolV', 'FPV', 'FPS', 'BitVec', 'StringV'}:
             self._op_raw[o] = getattr(self, '_op_raw_' + o)
         self._op_raw['Xor'] = self._op_raw_Xor
@@ -483,22 +485,7 @@ class BackendZ3(Backend):
 
         elif op_name == 'UNINTERPRETED':
             mystery_name = z3.Z3_get_symbol_string(ctx, z3.Z3_get_decl_name(ctx, decl))
-            args = [ ]
-
-            #
-            # TODO: DEPRECATED: remove the following after some reasonable amount of time.
-            #
-
-            if mystery_name == 'bvsmod_i':
-                l.error("Your Z3 is out of date. Please update angr-only-z3-custom or future releases of claripy will fail.")
-                op_name = '__mod__'
-                decl_num = z3.Z3_OP_BSMOD
-            elif mystery_name == 'bvsdiv_i':
-                l.error("Your Z3 is out of date. Please update angr-only-z3-custom or future releases of claripy will fail.")
-                op_name = 'SDiv'
-                decl_num = z3.Z3_OP_BSDIV
-            else:
-                l.error("Mystery operation %s in BackendZ3._abstract_internal. Please report this.", mystery_name)
+            l.error("Mystery operation %s in BackendZ3._abstract_internal. Please report this.", mystery_name)
         elif op_name == 'Extract':
             hi = z3.Z3_get_decl_int_parameter(ctx, decl, 0)
             lo = z3.Z3_get_decl_int_parameter(ctx, decl, 1)
@@ -562,20 +549,67 @@ class BackendZ3(Backend):
         op_name = op_map[z3_op_nums[decl_num]]
 
         if op_name == 'BitVecVal':
-            if z3.Z3_get_numeral_uint64(ctx, ast, self._c_uint64_p):
-                return self._c_uint64_p.contents.value
-            else:
-                bv_num = int(z3.Z3_get_numeral_string(ctx, ast))
-                return bv_num
+            return self._abstract_bv_val(ctx, ast)
         elif op_name == 'True':
             return True
         elif op_name == 'False':
             return False
         elif op_name in ('FPVal', 'MinusZero', 'MinusInf', 'PlusZero', 'PlusInf', 'NaN'):
             return self._abstract_fp_val(ctx, ast, op_name)
+        elif op_name == 'Concat':
+            # Quirk in how z3 might handle NaN encodings - it will not give us a fully evaluated model
+            # https://github.com/Z3Prover/z3/issues/518
+            # this case will be triggered if the z3 rewriter.hi_fp_unspecified is set to true
+            # update 6 jan 2020: idk if this is true anymore
+            nargs = z3.Z3_get_app_num_args(ctx, ast)
+            res = 0
+            for i in range(nargs):
+                arg_ast = z3.Z3_get_app_arg(ctx, ast, i)
+                arg_decl = z3.Z3_get_app_decl(ctx, arg_ast)
+                arg_decl_num = z3.Z3_get_decl_kind(ctx, arg_decl)
+                arg_size = z3.Z3_get_bv_sort_size(ctx, z3.Z3_get_sort(ctx, arg_ast))
+
+                neg = False
+                if arg_decl_num == z3.Z3_OP_BNEG:
+                    arg_ast = z3.Z3_get_app_arg(ctx, arg_ast, 0)
+                    arg_decl = z3.Z3_get_app_decl(ctx, arg_ast)
+                    arg_decl_num = z3.Z3_get_decl_kind(ctx, arg_decl)
+                    neg = True
+                if arg_decl_num != z3.Z3_OP_BNUM:
+                    raise BackendError("Weird z3 model")
+
+                arg_int = self._abstract_bv_val(ctx, arg_ast)
+                if neg:
+                    arg_int = (1<<arg_size)-arg_int
+                res <<= arg_size
+                res |= arg_int
+            return res
+        elif op_name == 'fpToIEEEBV':
+            # Another quirk in the way z3 might handle nan encodings. see above
+            # this case will be triggered if the z3 rewriter.hi_fp_unspecified is set to false
+            # update 6 jan 2020: idk if this is true anymore
+            arg_ast = z3.Z3_get_app_arg(ctx, ast, 0)
+            return self._abstract_fp_encoded_val(ctx, arg_ast)
+        elif op_name == 'INTERNAL' and self._abstract_string_check(ctx, ast):
+            # Quirk in how z3 encodes string constants
+            try:
+                assert(z3.Z3_get_decl_num_parameters(ctx, decl) == 1)
+                assert(z3.Z3_get_decl_parameter_kind(ctx, decl, 0) == z3.Z3_PK_SYMBOL)
+                symb = z3.Z3_get_decl_symbol_parameter(ctx, decl, 0)
+                assert(z3.Z3_get_symbol_kind(ctx, symb) == z3.Z3_STRING_SYMBOL)
+                return z3.Z3_get_symbol_string(ctx, symb).encode().decode('unicode_escape')
+                # see https://stackoverflow.com/a/58829514/3154996
+            except AssertionError:
+                raise BackendError("Weird z3 model")
         else:
             raise BackendError("Unable to abstract Z3 object to primitive")
 
+    def _abstract_bv_val(self, ctx, ast):
+        if z3.Z3_get_numeral_uint64(ctx, ast, self._c_uint64_p):
+            return self._c_uint64_p.contents.value
+        else:
+            return int(z3.Z3_get_numeral_string(ctx, ast))
+
     def _abstract_fp_val(self, ctx, ast, op_name):
         if op_name == 'FPVal':
             # TODO: do better than this
@@ -599,6 +633,64 @@ class BackendZ3(Backend):
         else:
             raise BackendError("Called _abstract_fp_val with unknown type")
 
+    def _abstract_fp_encoded_val(self, ctx, ast):
+        decl = z3.Z3_get_app_decl(ctx, ast)
+        decl_num = z3.Z3_get_decl_kind(ctx, decl)
+        op_name = op_map[z3_op_nums[decl_num]]
+        sort = z3.Z3_get_sort(ctx, ast)
+        ebits = z3.Z3_fpa_get_ebits(ctx, sort)
+        sbits = z3.Z3_fpa_get_sbits(ctx, sort) - 1  # includes sign bit
+
+        if op_name == 'FPVal':
+            # TODO: do better than this
+            fp_mantissa = int(z3.Z3_fpa_get_numeral_significand_string(ctx, ast))
+            fp_exp = int(z3.Z3_fpa_get_numeral_exponent_string(ctx, ast, True))
+            fp_sign_c = ctypes.c_int()
+            z3.Z3_fpa_get_numeral_sign(ctx, ast, ctypes.byref(fp_sign_c))
+            fp_sign = 1 if fp_sign_c.value != 0 else 0
+        elif op_name == 'MinusZero':
+            fp_sign = 1
+            fp_exp = 0
+            fp_mantissa = 0
+        elif op_name == 'MinusInf':
+            fp_sign = 1
+            fp_exp = (1<<ebits) - 1
+            fp_mantissa = 0
+        elif op_name == 'PlusZero':
+            fp_sign = 0
+            fp_exp = 0
+            fp_mantissa = 0
+        elif op_name == 'PlusInf':
+            fp_sign = 0
+            fp_exp = (1<<ebits) - 1
+            fp_mantissa = 0
+        elif op_name == 'NaN':
+            fp_sign = 0
+            fp_exp = (1<<ebits) - 1
+            fp_mantissa = 1
+        else:
+            raise BackendError("Called _abstract_fp_val with unknown type")
+
+        value = (fp_sign << (ebits + sbits)) | (fp_exp << sbits) | fp_mantissa
+        return value
+
+    def _abstract_string_check(self, ctx, ast):
+        # Check whether ast encodes a string constant
+        ast_sort = z3.Z3_get_sort(ctx, ast)
+        ast_sort_kind = z3.Z3_get_sort_kind(ctx, ast_sort)
+        if ast_sort_kind == z3.Z3_SEQ_SORT:
+            # At this point we know ast encodes some kind of sequence
+            seq_basis_sort = z3.Z3_get_seq_sort_basis(ctx, ast_sort)
+            seq_basis_sort_kind = z3.Z3_get_sort_kind(ctx, seq_basis_sort)
+            if seq_basis_sort_kind == z3.Z3_BV_SORT:
+                # At this point we know ast encodes a sequence of BVs
+                seq_basis_width = z3.Z3_get_bv_sort_size(ctx, seq_basis_sort)
+                if seq_basis_width == 8:
+                    # At this point we know ast encodes a sequence of 8-bit BVs
+                    # so we have a string! (per z3's definition of a string)
+                    return True
+        return False
+
     def solver(self, timeout=None):
         if not self.reuse_z3_solver or getattr(self._tls, 'solver', None) is None:
             s = z3.Solver(ctx=self._context)
@@ -633,6 +725,15 @@ class BackendZ3(Backend):
         else:
             s.add(*c)
 
+    def add(self, s, c, track=False):
+        converted = self.convert_list(c)
+        if track:
+            for a, nice_ast in zip(c, converted):
+                ast = nice_ast.ast
+                h = self._z3_ast_hash(ast)
+                self._ast_cache[h] = (a, ast)
+        return self._add(s, converted, track=track)
+
     def _unsat_core(self, s):
         cores = s.unsat_core()
         constraints = [ ]
@@ -1047,6 +1148,14 @@ class BackendZ3(Backend):
     def _op_raw_fpEQ(self, a, b):
         return z3.BoolRef(z3.Z3_mk_fpa_eq(self._context.ref(), a.as_ast(), b.as_ast()), self._context)
 
+    @condom
+    def _op_raw_fpIsNaN(self, a):
+        return z3.BoolRef(z3.Z3_mk_fpa_is_nan(self._context.ref(), a.as_ast()), self._context)
+
+    @condom
+    def _op_raw_fpIsInf(self, a):
+        return z3.BoolRef(z3.Z3_mk_fpa_is_inf(self._context.ref(), a.as_ast()), self._context)
+
     @condom
     def _op_raw_fpFP(self, sgn, exp, sig):
         return z3.FPRef(z3.Z3_mk_fpa_fp(self._context.ref(), sgn.ast, exp.ast, sig.ast), self._context)
@@ -1196,6 +1305,74 @@ class BackendZ3(Backend):
     def _identical(self, a, b):
         return a.eq(b)
 
+    # String operations:
+
+    @staticmethod
+    @condom
+    def _op_raw_StrConcat(*args):
+        return z3.Concat(*args)
+
+    @staticmethod
+    @condom
+    def _op_raw_StrSubstr(start_idx, count, initial_string):
+        return z3.SubString(
+            initial_string,
+            z3.BV2Int(start_idx),
+            z3.BV2Int(count)
+        )
+
+    @staticmethod
+    @condom
+    def _op_raw_StrExtract(high, low, str_val):
+        return z3.SubString(str_val, low, high + 1 - low)
+
+    @staticmethod
+    @condom
+    def _op_raw_StrLen(input_string, bitlength):
+        return z3.Int2BV(z3.Length(input_string), bitlength)
+
+    @staticmethod
+    @condom
+    def _op_raw_StrReplace(input_string, target, replacement):
+        return z3.Replace(input_string, target, replacement)
+
+    @staticmethod
+    @condom
+    def _op_raw_StrContains(input_string, substring):
+        return z3.Contains(input_string, substring)
+
+    @staticmethod
+    @condom
+    def _op_raw_StrPrefixOf(prefix, input_string):
+        return z3.PrefixOf(prefix, input_string)
+
+    @staticmethod
+    @condom
+    def _op_raw_StrSuffixOf(suffix, input_string):
+        return z3.SuffixOf(suffix, input_string)
+
+    @staticmethod
+    @condom
+    def _op_raw_StrIndexOf(string, pattern, start_idx, bitlength):
+        return z3.Int2BV(
+            z3.IndexOf(string, pattern, z3.BV2Int(start_idx)),
+            bitlength
+        )
+
+    @staticmethod
+    @condom
+    def _op_raw_StrToInt(input_string, bitlength):
+        return z3.Int2BV(z3.StrToInt(input_string), bitlength)
+
+    @staticmethod
+    @condom
+    def _op_raw_IntToStr(input_bvv):
+        return z3.IntToStr(z3.BV2Int(input_bvv))
+
+    @staticmethod
+    @condom
+    def _op_raw_UnitStr(input_bvv):
+        return z3.Unit(input_bvv)
 #
 # this is for the actual->abstract conversion above
 #
@@ -1347,6 +1524,7 @@ op_map = {
     'Z3_OP_FPA_RM_TOWARD_POSITIVE': 'RM_RTP',
     'Z3_OP_FPA_RM_TOWARD_NEGATIVE': 'RM_RTN',
 
+    'Z3_OP_INTERNAL': 'INTERNAL',
     'Z3_OP_UNINTERPRETED': 'UNINTERPRETED'
 }
 
@@ -1355,7 +1533,7 @@ from ..ast.bv import BV, BVV
 from ..ast.bool import BoolV, Bool
 from ..ast.fp import FP, FPV
 from ..ast.strings import StringV, StringS
-from ..operations import backend_operations, backend_fp_operations
+from ..operations import backend_operations, backend_fp_operations, backend_strings_operations
 from ..fp import FSort, RM, RM_NearestTiesEven, RM_NearestTiesAwayFromZero, RM_TowardsPositiveInf, RM_TowardsNegativeInf, RM_TowardsZero
 from ..errors import ClaripyError, BackendError, ClaripyOperationError
 from .. import _all_operations
